//! Code for converting between different casing disciplines.

use proc_macro2::Ident;

pub(crate) trait CaseConvert {
    /// Converts a name in snake_case to PascalCase.
    fn snake_to_pascal(&self) -> String;
    /// Converts a name in PascalCase to snake_case.
    fn pascal_to_snake(&self) -> String;
}

impl CaseConvert for String {
    fn snake_to_pascal(&self) -> String {
        let mut pascal = String::with_capacity(self.len());
        let mut prev_underscore = true;
        for c in self.chars() {
            if '_' == c {
                prev_underscore = true;
            } else {
                if prev_underscore {
                    pascal.extend(c.to_uppercase());
                } else {
                    pascal.push(c);
                }
                prev_underscore = false;
            }
        }
        pascal
    }

    fn pascal_to_snake(&self) -> String {
        let mut snake = String::with_capacity(self.len());
        let mut prev_lower = false;
        for c in self.chars() {
            if c.is_uppercase() {
                if prev_lower {
                    snake.push('_');
                }
                snake.extend(c.to_lowercase());
                prev_lower = false;
            } else {
                prev_lower = true;
                snake.push(c);
            }
        }
        snake
    }
}

impl CaseConvert for Ident {
    fn snake_to_pascal(&self) -> String {
        self.to_string().snake_to_pascal()
    }

    fn pascal_to_snake(&self) -> String {
        self.to_string().pascal_to_snake()
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn string_snake_to_pascal_multiple_segments() {
        const EXPECTED: &str = "FirstSecondThird";
        let input = String::from("first_second_third");

        let actual = input.snake_to_pascal();

        assert_eq!(EXPECTED, actual);
    }

    #[test]
    fn string_snake_to_pascal_single_segment() {
        const EXPECTED: &str = "First";
        let input = String::from("first");

        let actual = input.snake_to_pascal();

        assert_eq!(EXPECTED, actual);
    }

    #[test]
    fn string_snake_to_pascal_empty_string() {
        const EXPECTED: &str = "";
        let input = String::from(EXPECTED);

        let actual = input.snake_to_pascal();

        assert_eq!(EXPECTED, actual);
    }

    #[test]
    fn string_snake_to_pascal_leading_underscore() {
        const EXPECTED: &str = "First";
        let input = String::from("_first");

        let actual = input.snake_to_pascal();

        assert_eq!(EXPECTED, actual);
    }

    #[test]
    fn string_snake_to_pascal_leading_underscores() {
        const EXPECTED: &str = "First";
        let input = String::from("__first");

        let actual = input.snake_to_pascal();

        assert_eq!(EXPECTED, actual);
    }

    #[test]
    fn string_snake_to_pascal_multiple_underscores() {
        const EXPECTED: &str = "FirstSecondThird";
        let input = String::from("first__second___third");

        let actual = input.snake_to_pascal();

        assert_eq!(EXPECTED, actual);
    }

    #[test]
    fn string_pascal_to_snake_multiple_segments() {
        const EXPECTED: &str = "first_second_third";
        let input = String::from("FirstSecondThird");

        let actual = input.pascal_to_snake();

        assert_eq!(EXPECTED, actual);
    }

    #[test]
    fn string_pascal_to_snake_single_segment() {
        let input = String::from("First");
        const EXPECTED: &str = "first";

        let actual = input.pascal_to_snake();

        assert_eq!(EXPECTED, actual);
    }

    #[test]
    fn string_pascal_to_snake_empty_string() {
        const EXPECTED: &str = "";
        let input = String::from(EXPECTED);

        let actual = input.pascal_to_snake();

        assert_eq!(EXPECTED, actual);
    }

    #[test]
    fn string_pascal_to_snake_consecutive_uppercase() {
        const EXPECTED: &str = "kernel_mc";
        let input = String::from("KernelMC");

        let actual = input.pascal_to_snake();

        assert_eq!(EXPECTED, actual);
    }
}
