//! Code to assist with error handling, primarily [MaybeErr].

/// A wrapper around `Option<syn::Error>` which allows errors to be easily combined and created
/// from iterators of [syn::Error].
#[derive(Debug, Default)]
pub(crate) struct MaybeErr(Option<syn::Error>);

impl MaybeErr {
    pub(crate) fn none() -> Self {
        MaybeErr(None)
    }

    pub(crate) fn combine(self, other: Self) -> Self {
        let option = match (self.0, other.0) {
            (Some(left), right) => fold_errs(right, left),
            (left, Some(right)) => fold_errs(left, right),
            _ => None,
        };
        MaybeErr(option)
    }
}

impl From<MaybeErr> for Result<(), syn::Error> {
    fn from(value: MaybeErr) -> Self {
        if let Some(err) = value.0 {
            Err(err)
        } else {
            Ok(())
        }
    }
}

impl From<syn::Error> for MaybeErr {
    fn from(err: syn::Error) -> Self {
        MaybeErr(Some(err))
    }
}

impl FromIterator<syn::Error> for MaybeErr {
    fn from_iter<T: IntoIterator<Item = syn::Error>>(iter: T) -> Self {
        MaybeErr(iter.into_iter().fold(None, fold_errs))
    }
}

fn fold_errs(accum: Option<syn::Error>, curr: syn::Error) -> Option<syn::Error> {
    if let Some(mut accum) = accum {
        accum.combine(curr);
        Some(accum)
    } else {
        Some(curr)
    }
}

#[cfg(test)]
/// Panics if the given value does not transform into `Ok`.
pub(crate) fn assert_ok<T, E: Into<syn::Result<T>>>(maybe_err: E) {
    let result: syn::Result<T> = maybe_err.into();
    assert!(result.is_ok(), "{}", result.err().unwrap());
}

#[cfg(test)]
/// Panics if the given value does not transform into `Err` with an error value containing the
/// given message.
pub(crate) fn assert_err<T, E: Into<syn::Result<T>>>(maybe_err: E, expected_msg: &str) {
    let result: syn::Result<T> = maybe_err.into();
    assert!(result.is_err());
    assert_eq!(expected_msg, result.err().unwrap().to_string());
}

/// User-visible compile time error messages.
pub(crate) mod msgs {
    /// Indicates that a duplicate transition has been defined.
    ///
    /// This means it is either handling the same message type in the same state sending the
    /// same message type in the same state as another transition.
    pub(crate) const DUPLICATE_TRANSITION: &str = "Duplicate transition.";
    pub(crate) const NO_MSG_SENT_OR_RECEIVED: &str = "A transition must send or receive a message.";
    pub(crate) const UNDECLARED_STATE: &str = "State was not declared.";
    pub(crate) const UNUSED_STATE: &str = "State was declared but never used.";
    pub(crate) const UNMATCHED_OUTGOING: &str = "No receiver found for message type.";
    pub(crate) const UNMATCHED_INCOMING: &str = "No sender found for message type.";
    pub(crate) const UNDELIVERABLE: &str =
        "Receiver must either be a service, an owned state, or an out state, or the message must be a reply.";
    pub(crate) const INVALID_REPLY: &str =
        "Replies can only be used in transitions which handle messages.";
    pub(crate) const MULTIPLE_REPLIES: &str =
        "Only a single reply can be sent in response to any message.";
    pub(crate) const UNOBSERVABLE_STATE: &str =
        "This client state is not allowed because it only receives replies. Such a state cannot be observed.";
    pub(crate) const CLIENT_USED_IN_SERVICE: &str =
        "A client state cannot be used inside `service()`.";
}
